-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved


THCore = {}
local THCore_mt = THUtils.createClass(THCore)
function THCore.initXMLSchema(xmlSchema, xmlPath)
end
function THCore.initXMLSchemaSettings(xmlSchema, xmlPath)
    THSettingsManager.initXMLSchemaSettings(xmlSchema, xmlPath)
end
function THCore.initXMLSchemaSavegame(xmlSchema, xmlPath)
end
function THCore.new(xmlDataKey, customMt)
    customMt = customMt or THCore_mt
    if THUtils.argIsValid(type(xmlDataKey) == "string" and xmlDataKey ~= "", "xmlDataKey", xmlDataKey) then
        local self = setmetatable({}, customMt)
        local missionInfo = g_mpLoadingScreen.missionInfo
        local missionDynamicInfo = g_mpLoadingScreen.missionDynamicInfo
        self.isServer = g_server ~= nil
        self.isClient = g_client ~= nil
        self.isMultiplayer = missionDynamicInfo.isMultiplayer
        self.isEnabled = true
        self.isInitFinished = false
        self.debugFlagId = THUtils.createDebugFlag(xmlDataKey)
        self.inputBinding = g_inputBinding
        self.messageCenter = g_messageCenter
        self.gameSettings = g_gameSettings
        self.taskManager = g_asyncTaskManager
        self.i18n = g_i18n
        self.gui = g_gui
        self.fruitTypeManager = g_fruitTypeManager
        self.fillTypeManager = g_fillTypeManager
        self.farmlandManager = g_farmlandManager
        self.fieldManager = g_fieldManager
        self.animationManager = g_animationManager
        self.effectManager = g_effectManager
        self.soundManager = g_soundManager
        self.actionEventIds = {}
        self.registeredActions = {}
        self.delayedTaskFunctions = {}
        self.modList = missionDynamicInfo.mods
        self.modInfo = nil -- set by initLoadedMods()
        self.modName = g_currentModName
        self.modPath = g_currentModDirectory
        self.modTitle = ""
        self.modVersion = ""
        self.baseHUDFilename = g_baseHUDFilename
        self.baseUIFilename = g_baseUIFilename
        self.sdkPath = self.modPath .. "sdk/"
        self.xmlDataKey = xmlDataKey
        self.modSettingsXMLPath = g_currentModSettingsDirectory
        self.modSettingsXMLFilename = self.modSettingsXMLPath .. xmlDataKey .. ".xml"
        createFolder(self.modSettingsXMLPath)
        self.savegameXMLFilename = "/" .. xmlDataKey .. ".xml"
        self.loadedMods = {
            byId = {},
            byIndex = {},
            byTarget = {}
        }
        self.mapSettingsModName = "FS22_TH_MapSettings"
        self.mapSettingsModInfo = nil -- set my initLoadedMods()
        local mapInfo = missionInfo.map
        local mapName = mapInfo.id
        if mapInfo.isModMap then
            if THUtils.getNoNil(mapInfo.customEnvironment, "") ~= "" then
                mapName = mapInfo.customEnvironment
            end
        end
        self.mapSettingsXMLPath = self.modSettingsXMLPath .. "maps/" .. mapName .. "/"
        self.mapInfo = mapInfo
        self.mapName = mapName
        self.mapTitle = ""
        self.mapVersion = ""
        self.mapIsMod = mapInfo.isModMap
        self.mapXMLFilename = mapInfo.mapXMLFilename
        self.mapModInfo = nil -- set by initLoadedMods()
        if self.mapIsMod then
            self.mapPath = mapInfo.baseDirectory
        end
        self.eventListeners = {}
        THUtils.setFunctionHook("XMLManager", "initSchemas", THCore, nil, self)
        return self
    end
end
function THCore:loadModSettings()
    local loadSuccess = true
    local xmlSchema = self.xmlSchemaSettings
    local xmlDataKey = self.xmlDataKey
    local modSettingsFilename = self.modSettingsXMLFilename
    if xmlSchema ~= nil then
        local modSettingsXMLFile = XMLFile.loadIfExists(xmlDataKey .. "ModSettingsXML", modSettingsFilename, xmlSchema)
        if modSettingsXMLFile ~= nil then
            loadSuccess = self:loadModSettingsFromXMLFile(modSettingsXMLFile, xmlDataKey)
            modSettingsXMLFile:delete()
        end
    end
    return loadSuccess
end
function THCore:loadModSettingsFromXMLFile(xmlFile, xmlKey)
    return true
end
function THCore:saveModSettings()
    local saveSuccess = true
    local xmlSchema = self.xmlSchema
    local xmlDataKey = self.xmlDataKey
    local modSettingsFilename = self.modSettingsXMLFilename
    if xmlSchema ~= nil then
        local modSettingsXMLFile = nil -- placeholder
        if fileExists(modSettingsFilename) then
            modSettingsXMLFile = XMLFile.load(xmlDataKey .. "ModSettingsXML", modSettingsFilename, xmlSchema)
        else
            modSettingsXMLFile = XMLFile.create(xmlDataKey .. "ModSettingsXML", modSettingsFilename, xmlDataKey, xmlSchema)
        end
        if modSettingsXMLFile ~= nil then
            saveSuccess = self:saveModSettingsToXMLFile(modSettingsXMLFile, xmlDataKey)
            if saveSuccess then
                modSettingsXMLFile:save()
            end
            modSettingsXMLFile:delete()
        else
            saveSuccess = false
        end
    end
    return saveSuccess
end
function THCore:saveModSettingsToXMLFile(xmlFile, xmlKey)
    return true
end
function THCore:getMapXMLFile(dataKey, xmlSchema, useSettingsMod)
    local baseXMLKey = self.xmlDataKey
    local mapXMLName = string.format("%sConfigXML", dataKey)
    if useSettingsMod ~= false and self.mapSettingsModInfo ~= nil then
        local mapPath = self.mapSettingsModInfo.path
        local mapXMLFilename = string.format("%smap/%s/%s.xml", mapPath, self.mapName, dataKey)
        local mapXMLFile = XMLFile.loadIfExists(mapXMLName, mapXMLFilename, xmlSchema)
        if mapXMLFile == nil then
            mapXMLFilename = string.format("%smap/%s.xml", mapPath, dataKey) -- look for global config
            mapXMLFile = XMLFile.loadIfExists(mapXMLName, mapXMLFilename, xmlSchema)
        end
        if mapXMLFile ~= nil then
            return mapXMLFile
        end
    end
    if self.mapIsMod and self.mapModInfo ~= nil then
        local mapXMLFilename = Utils.getFilename(self.mapXMLFilename, self.mapPath)
        local mapXMLFile = XMLFile.loadIfExists(dataKey .. "FilenameXML", mapXMLFilename)
        if mapXMLFile ~= nil then
            local filenameKey = string.format("map.%s.%s#filename", baseXMLKey, dataKey)
            local dataXMLFilename = mapXMLFile:getString(filenameKey)
            if dataXMLFilename ~= nil then
                local dataXMLFile = XMLFile.loadIfExists(mapXMLName, dataXMLFilename, xmlSchema)
                if dataXMLFile ~= nil then
                    return dataXMLFile
                end
            end
        end
    end
end
function THCore:loadFromSavegame(savegame)
    local xmlSchema = self.xmlSchemaSavegame
    local xmlDataKey = self.xmlDataKey
    local savegameDirectory = savegame.savegameDirectory
    if xmlSchema ~= nil and savegameDirectory ~= nil then
        THUtils.call(function()
            local savegameXMLFilename = savegameDirectory .. self.savegameXMLFilename
            local savegameXMLFile = XMLFile.loadIfExists(xmlDataKey .. "SavegameXML", savegameXMLFilename, xmlSchema)
            if savegameXMLFile ~= nil then
                self:loadFromSavegameXMLFile(savegame, savegameXMLFile, xmlDataKey)
                savegameXMLFile:delete()
            end
        end)
    end
end
function THCore:loadFromSavegameXMLFile(savegame, xmlFile, xmlKey)
    return true
end
function THCore:saveToSavegame(savegame)
    local xmlSchema = self.xmlSchemaSavegame
    local xmlDataKey = self.xmlDataKey
    local savegameDirectory = savegame.savegameDirectory
    if xmlSchema ~= nil and savegameDirectory ~= nil then
        THUtils.call(function()
            local savegameXMLFilename = savegameDirectory .. self.savegameXMLFilename
            local savegameXMLFile = XMLFile.create(xmlDataKey .. "SavegameXML", savegameXMLFilename, xmlDataKey, xmlSchema)
            if savegameXMLFile ~= nil then
                if self:saveToSavegameXMLFile(savegame, savegameXMLFile, xmlDataKey) then
                    savegameXMLFile:save()
                end
                savegameXMLFile:delete()
            end
        end)
    end
end
function THCore:saveToSavegameXMLFile(savegame, xmlFile, xmlKey)
    return true
end
function THCore:onInitialize()
    if not self.isInitFinished then
        self:initLoadedMods()
        self.networkUpdater = THNetworkUpdater.new(self.isServer, self.isClient, nil, self)
        self.networkUpdater:register(true)
        self.dirtyFlag = self:getNextDirtyFlag()
        self.settingsManager = THSettingsManager.new()
        self:addConsoleCommand("setDebugFlag", "Sets specified debug flag", "consoleCommandSetDebugFlag")
        self:addConsoleCommand("setDebugLevel", "Sets specified debug level", "consoleCommandSetDebugLevel")
        THUtils.setFunctionHook("Player", "readStream", THCore, "hook_initialReadStream", self)
        THUtils.setFunctionHook("Player", "writeStream", THCore, "hook_initialWriteStream", self)
        THUtils.setFunctionHook("Server", "sendEventIds", THCore, nil, self)
        THUtils.setFunctionHook("Mission00", "setMissionInfo", THCore, nil, self)
        THUtils.setFunctionHook("Mission00", "loadEnvironment", THCore, nil, self)
        THUtils.setFunctionHook("FSBaseMission", "loadMapFinished", THCore, nil, self)
        THUtils.setFunctionHook("BaseMission", "registerActionEvents", THCore, nil, self)
        THUtils.setFunctionHook("BaseMission", "unregisterActionEvents", THCore, nil, self)
        THUtils.setFunctionHook("FSCareerMissionInfo", "saveToXMLFile", THCore, nil, self)
        self.isInitFinished = true
    end
end
function THCore:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
    self.mission = mission
    self.missionInfo = missionInfo
    self.missionDynamicInfo = missionDynamicInfo
    self.savegameController = mission.savegameController
    self.savegameIndex = missionInfo.savegameIndex
    if self.isServer and self.savegameIndex ~= nil then
        self.savegame = self.savegameController:getSavegame(self.savegameIndex)
    end
end
function THCore:onLoadEnvironment(environment, mission)
end
function THCore:onSendEventIds(connection)
end
function THCore:onSendObjectIds(connection)
end
function THCore:onLoadMapFinished(mission)
    if self.isServer then
        self:raiseActive()
    end
end
function THCore:onRegisterActionEvents(mission)
end
function THCore:onUnregisterActionEvents(mission)
end
function THCore:onUpdate(dt)
end
function THCore:onUpdateTick(dt)
end
function THCore:onReadStream(streamId, connection)
end
function THCore:onWriteStream(streamId, connection)
end
function THCore:onReadUpdateStream(streamId, timeStamp, connection)
end
function THCore:onWriteUpdateStream(streamId, connection, dirtyMask)
end
function THCore:addConsoleCommand(name, desc, callbackName)
    local modName = self.modName
    addConsoleCommand(modName:gsub("^FS22_TH_", "th") .. ":" .. name, desc, callbackName, self)
end
function THCore:consoleCommandSetDebugFlag(flagId, isEnabled, sendToServer)
    local modName = self.modName
    local usageMsg = "Usage: " .. modName:gsub("^FS22_TH_", "th") .. ":setDebugFlag flagId isEnabled sendToServer"
    isEnabled = THUtils.toBoolean(isEnabled)
    sendToServer = THUtils.toBoolean(sendToServer)
    local function showUsageMessage()
        printf(usageMsg)
        printf("")
        printf("Allowed debug flagIds:")
        for otherFlagId in pairs(THUtils.debugFlags) do
            printf("- %s", otherFlagId)
        end
        printf("")
    end
    if isEnabled == nil then
        showUsageMessage()
    else
        local realFlagId, newIsEnabled = THUtils.setDebugFlag(flagId, isEnabled)
        local enabledText = nil -- placeholder
        if newIsEnabled == true then
            enabledText = "enabled"
        elseif newIsEnabled == false then
            enabledText = "disabled"
        end
        if enabledText == nil then
            showUsageMessage()
        else
            if realFlagId == nil then
                THUtils.displayMsg("Debugger is %s", enabledText)
            else
                THUtils.displayMsg("Debugging for %q is %s", realFlagId, enabledText)
            end
            if not self.isServer and sendToServer == true then
                THSetDebugFlagEvent.sendEvent(realFlagId, newIsEnabled)
            end
        end
    end
end
function THCore:consoleCommandSetDebugLevel(debugLevelId, sendToServer)
    local modName = self.modName
    local usageMsg = "Usage: " .. modName:gsub("^FS22_TH_", "th") .. ":setDebugLevel debugLevelId"
    sendToServer = THUtils.toBoolean(sendToServer)
    if debugLevelId == nil or debugLevelId == "" then
        printf(usageMsg)
    else
        THUtils.setDebugLevel(debugLevelId)
        if not self.isServer and sendToServer == true then
            THSetDebugLevelEvent.sendEvent(debugLevelId)
        end
    end
end
function THCore:addEventListener(target)
    if THUtils.argIsValid(type(target) == "table", "target", target, true) then -- ! stop on error
        local isTargetFound = false
        for _, otherTarget in pairs(self.eventListeners) do
            if target == otherTarget then
                isTargetFound = true
                break
            end
        end
        if not isTargetFound then
            table.insert(self.eventListeners, target)
        end
    end
end
function THCore:removeEventListener(target)
    local numEventListeners = #self.eventListeners
    if target ~= nil and numEventListeners > 0 then
        local listenerIndex = 1
        while true do
            if self.eventListeners[listenerIndex] == nil then
                break
            end
            if self.eventListeners[listenerIndex] == target then
                table.remove(self.eventListeners, listenerIndex)
            else
                listenerIndex = listenerIndex + 1
            end
        end
    end
end
function THCore:raiseEvent(eventName, ...)
    local numEventListeners = #self.eventListeners
    if numEventListeners > 0 then
        for listenerIndex = 1, numEventListeners do
            local listener = self.eventListeners[listenerIndex]
            if listener[eventName] ~= nil then
                listener[eventName](listener, ...)
            end
        end
    end
end
function THCore:getNextDirtyFlag()
    return self.networkUpdater:getNextDirtyFlag()
end
function THCore:raiseDirtyFlags(dirtyFlag)
    if dirtyFlag ~= nil then
        return self.networkUpdater:raiseDirtyFlags(dirtyFlag)
    end
end
function THCore:raiseActive()
    self.networkUpdater:raiseActive()
end
function THCore:addDelayedLoadTask(taskFunc)
    if THUtils.argIsValid(type(taskFunc) == "function", "taskFunc", taskFunc, true) then -- ! stop on error
        table.insert(self.delayedTaskFunctions, taskFunc)
    end
end
function THCore:runDelayedLoadTasks()
    local numTasks = #self.delayedTaskFunctions
    if numTasks > 0 then
        for taskIndex = numTasks, 1, -1 do
            local taskFunc = self.delayedTaskFunctions[taskIndex]
            THUtils.call(taskFunc)
        end
        THUtils.clearTable(self.delayedTaskFunctions)
    end
end
function THCore:initLoadedMods()
    local globalEnv = THUtils.getGlobalEnvironment()
    for _, modInfo in ipairs(self.modList) do
        local modName = THUtils.getNoNil(modInfo.modName, "")
        if modName ~= "" then
            local modEntry = {
                id = modName:upper(),
                name = modName,
                info = modInfo,
                path = modInfo.modDir,
                file = modInfo.modFile,
                index = #self.loadedMods.byIndex + 1,
                title = modInfo.title,
                version = modInfo.version,
                environment = globalEnv[modName]
            }
            if type(modEntry.version) == "string" then
                local numericVersionText = string.gsub(modEntry.version, "[^0-9]", "")
                modEntry.numericVersion = tonumber(numericVersionText)
            end
            modEntry.numericVersion = THUtils.getNoNil(modEntry.numericVersion, 1)
            self.loadedMods.byId[modEntry.id] = modEntry
            self.loadedMods.byIndex[modEntry.index] = modEntry
            self.loadedMods.byTarget[modEntry] = modEntry
            if modName == self.modName then
                self.modInfo = modEntry
                self.modTitle = modEntry.title
                self.modVersion = modEntry.version
            elseif modName == self.mapSettingsModName then
                self.mapSettingsModInfo = modEntry
            end
            if self.mapIsMod and self.mapName ~= nil and modName == self.mapName then
                self.mapModInfo = modEntry
                self.mapTitle = modEntry.title
                self.mapVersion = modEntry.version
            end
        end
    end
end
function THCore:getLoadedMod(modName, verbose, raise)
    local modEntry = nil -- placeholder
    if type(modName) == "string" then
        modEntry = self.loadedMods.byId[modName:upper()]
    elseif modName ~= nil then
        verbose = true
        raise = THUtils.getNoNil(raise, true)
    end
    THUtils.msgOnTrue(modEntry == nil and verbose == true, raise, THUtils.MSG.ARGUMENT_INVALID, "modName", modName)
    return modEntry
end
function THCore:getLoadedModByClassName(className, modName)
    local modEntry = nil -- placeholder
    if modName ~= nil then
        modEntry = self:getLoadedMod(modName)
        if modEntry == nil then
            return
        end
    end
    local function getModClass(pModEntry)
        if pModEntry ~= nil and pModEntry.environment ~= nil then
            local modClass = pModEntry.environment[className]
            if type(modClass) == "table" then
                return modClass
            end
        end
    end
    if modEntry == nil then
        local loadedModsArray, numLoadedMods = self:getLoadedModList(true)
        if numLoadedMods > 0 then
            for i = 1, numLoadedMods do
                modEntry = loadedModsArray[i]
                local modClass = getModClass(modEntry)
                if modClass ~= nil then
                    return modEntry, modClass
                end
            end
        end
    else
        local modClass = getModClass(modEntry)
        if modClass ~= nil then
            return modEntry, modClass
        end
    end
end
function THCore:getLoadedModList(byIndex)
    local numLoadedMods = #self.loadedMods.byIndex
    if byIndex == true then
        return self.loadedMods.byIndex, numLoadedMods
    else
        return self.loadedMods.byId, numLoadedMods
    end
end
function THCore:getModVersion(modName, getNumeric)
    local modEntry = self.modInfo
    if modName ~= nil then
        modEntry = self:getLoadedMod(modName)
    end
    if modEntry ~= nil then
        if getNumeric ~= false then
            return modEntry.version
        else
            return modEntry.numericVersion
        end
    end
end
function THCore:hook_initSchemas(superFunc, xmlManager, ...)
    local function appendFunc(...)
        self.taskManager:addSubtask(function()
            if not self.isInitFinished then
                self:onInitialize()
                self:raiseEvent("onInitialize")
                self:runDelayedLoadTasks()
                self.isInitFinished = true
            end
        end)
        return ...
    end
    return appendFunc(superFunc(xmlManager, ...))
end
function THCore:hook_setMissionInfo(superFunc, mission, missionInfo, missionDynamicInfo, ...)
    local function appendFunc(...)
        self.taskManager:addTask(function()
            self:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
            self:raiseEvent("onSetMissionInfo", mission, missionInfo, missionDynamicInfo)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, missionInfo, missionDynamicInfo, ...))
end
function THCore:hook_loadEnvironment(superFunc, mission, ...)
    local function appendFunc(...)
        THUtils.call(function()
            self:onLoadEnvironment(mission.environment, mission)
            self:raiseEvent("onLoadEnvironment", mission.environment, mission)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THCore:hook_sendEventIds(superFunc, server, connection, ...)
    local function appendFunc(...)
        THUtils.call(function()
            self:onSendEventIds(connection)
            self:raiseEvent("onSendEventIds", connection)
        end)
        return ...
    end
    return appendFunc(superFunc(server, connection, ...))
end
function THCore:hook_initialReadStream(superFunc, player, streamId, connection, ...)
    local function appendFunc(...)
        if connection:getIsServer() then -- client reading from server
            local networkUpdaterData, networkUpdater = THUtils.getDataTable(self.networkUpdater)
            if streamReadBool(streamId) then -- needs registration
                local networkUpdaterId = NetworkUtil.readNodeObjectId(streamId)
                networkUpdater:readStream(streamId, connection)
                g_client:finishRegisterObject(networkUpdater, networkUpdaterId)
                networkUpdaterData.isRegisteredInStream = true
            end
        end
        return ...
    end
    return appendFunc(superFunc(player, streamId, connection, ...))
end
function THCore:hook_initialWriteStream(superFunc, player, streamId, connection, ...)
    local function appendFunc(...)
        if not connection:getIsServer() then -- server writing to client
            local networkUpdaterData, networkUpdater = THUtils.getDataTable(self.networkUpdater)
            if streamWriteBool(streamId, networkUpdaterData.isRegisteredInStream ~= true) then
                local networkUpdaterId = NetworkUtil.getObjectId(networkUpdater)
                NetworkUtil.writeNodeObjectId(streamId, networkUpdaterId)
                networkUpdater:writeStream(streamId, connection)
                g_server:registerObjectInStream(connection, networkUpdater)
                networkUpdaterData.isRegisteredInStream = true
            end
        end
        return ...
    end
    return appendFunc(superFunc(player, streamId, connection, ...))
end
function THCore:hook_loadMapFinished(superFunc, mission, ...)
    local function appendFunc(...)
        THUtils.call(function()
            if not mission.cancelLoading then
                self:onLoadMapFinished(mission)
                self:raiseEvent("onLoadMapFinished", mission)
                if self.isServer and self.savegame ~= nil then
                    self:loadFromSavegame(self.savegame)
                    self:raiseEvent("loadFromSavegame", self.savegame)
                end
            end
        end)
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THCore:hook_registerActionEvents(superFunc, mission, ...)
    local function appendFunc(...)
        THUtils.call(function()
            self:onRegisterActionEvents(mission)
            self:raiseEvent("onRegisterActionEvents", mission)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THCore:hook_unregisterActionEvents(superFunc, mission, ...)
    local function appendFunc(...)
        THUtils.call(function()
            self:onUnregisterActionEvents(mission)
            self:raiseEvent("onUnregisterActionEvents", mission)
        end)
        return ...
    end
    return appendFunc(superFunc(mission, ...))
end
function THCore:hook_saveToXMLFile(superFunc, missionInfo, ...)
    local function appendFunc(...)
        THUtils.call(function()
            if self.isServer and self.savegame ~= nil then
                self:saveToSavegame(self.savegame)
                self:raiseEvent("saveToSavegame", self.savegame)
            end
        end)
        return ...
    end
    return appendFunc(superFunc(missionInfo, ...))
end