-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved


THUtils = {}
THUtils.MOD_NAME = g_currentModName
THUtils.DATA_KEY = string.format("%sData", THUtils.MOD_NAME)
THUtils.DEBUG_LEVEL = {
    ERROR = 1,
    WARNING = 2,
    NORMAL = 3,
    UPDATE = 4,
    LOOP = 5
}
THUtils.debugFlags = {}
THUtils.globalDebugFlag = false
THUtils.debugLevel = THUtils.DEBUG_LEVEL.NORMAL
THUtils.functionCache = {}
THUtils.MSG = {
    INTERNAL_ERROR = "An internal error has occurred",
    ARGUMENT_INVALID = "Invalid argument %q (%s) in function call"
}
THUtils.XMLMSG = {
    DUPLICATE_ENTRY = "duplicate entry found",
    INVALID_VALUE = "missing or invalid value",
    INVALID_FILL_TYPE = "missing or invalid fillType",
    INVALID_FRUIT_TYPE = "missing or invalid fruitType",
    VALUE_OUT_OF_RANGE = "must be between %s and %s",
    VALUE_GREATER_ZERO = "must be greater than zero",
    VALUE_GREATER_EQUAL_ZERO = "must be greater than or equal to zero",
    VALUE_LESS_EQUAL_AMOUNT = "must be less than or equal to %s"
}
function THUtils.displayMsg(text, ...)
    if text == "" then
        print(text)
    else
        if text == nil or type(text) ~= "string" then
            text = string.format("ERROR: Invalid displayMsg text, %q", text)
        else
            if select("#", ...) > 0 then
                text = string.format(text, ...)
            end
        end
        local modName = THUtils.MOD_NAME
        local msgString = string.format("** [%s]: %s", modName, text)
        print(msgString)
    end
end
function THUtils.warningMsg(text, ...)
    if type(text) == "string" then
        text = "WARNING: " .. text
    end
    THUtils.displayMsg(text, ...)
end
function THUtils.errorMsg(errorLevel, text, ...)
    if errorLevel == true then errorLevel = 2 end -- switch to numeric format
    if type(errorLevel) == "number" or errorLevel == nil or errorLevel == false then
        if type(text) == "string" then
            text = "ERROR: " .. text
        end
        local errorText = text
        if select("#", ...) > 0 then
            errorText = string.format(text, ...)
        end
        if errorLevel == nil or errorLevel == false then
            THUtils.displayMsg(errorText)
            if errorLevel == nil then
                printCallstack()
            end
        else
            error(errorText, errorLevel)
        end
    else
        local errorText = string.format(THUtils.MSG.ARGUMENT_INVALID, "errorLevel", errorLevel)
        error(errorText, 1)
    end
end
function THUtils.assert(expression, raise, text, ...)
    if not expression then
        if raise ~= nil then
            if raise == true then
                raise = 3 -- increased to include this function call
            elseif type(raise) == "number" then
                raise = raise + 1 -- increased to include this function call
            end
        end
        if raise == false then
            THUtils.displayMsg(text, ...)
        else
            THUtils.errorMsg(raise, text, ...)
        end
    end
    return expression
end
function THUtils.msgOnTrue(expression, raise, text, ...)
    if expression then
        if raise ~= nil then
            if raise == true then
                raise = 3 -- increased to include this function call
            elseif type(raise) == "number" then
                raise = raise + 1 -- increased to include this function call
            end
        end
        if raise == false then
            THUtils.displayMsg(text, ...)
        else
            THUtils.errorMsg(raise, text, ...)
        end
        return true
    end
    return false
end
function THUtils.argIsValid(expression, argName, argValue, raise)
    if not expression then
        if raise ~= nil then
            if raise == false then
                raise = nil -- show callstack only
            elseif raise == true then
                raise = 3 -- increased to include this function call
            elseif type(raise) == "number" then
                raise = raise + 1 -- increase to include this function call
            end
        end
        THUtils.errorMsg(raise, THUtils.MSG.ARGUMENT_INVALID, argName, argValue)
        return false
    end
    return true
end
function THUtils.xmlLoadMsg(xmlKey, attrName, attrValue, isError, text, ...)
    local xmlText = "XML Warning: %s"
    if isError == true then
        xmlText = "XML Error: %s"
    end
    THUtils.displayMsg(xmlText, xmlKey)
    if attrName == nil then
        THUtils.displayMsg(">> " .. text, ...)
    else
        local newText = string.format("Attribute (%s=%q) ", attrName, attrValue)
        THUtils.displayMsg(">> " .. newText .. text, ...)
    end
end
function THUtils.printTable(tbl, ...)
    if g_thDebugger ~= nil then
        g_thDebugger:printTable(tbl, ...)
    else
        printf("Printing %s", tbl)
        if type(tbl) == "table" then
            for k, v in pairs(tbl) do
                printf("[%s] = %s", k, v)
            end
        else
            printf("- Invalid table")
        end
        printf("")
    end
end
function THUtils.debugPrintTable(flagId, tbl, ...)
    if THUtils.getIsDebugEnabled(flagId) then
        THUtils.printTable(tbl, ...)
    end
end
function THUtils.createDebugFlag(flagId)
    if type(flagId) == "table" and flagId.debugFlagId ~= nil then
        flagId = flagId.debugFlagId
    end
    if flagId == nil or flagId == "" then
        THUtils.errorMsg(nil, THUtils.MSG.ARGUMENT_INVALID, "flagId", flagId)
    else
        local realFlagId = tostring(flagId):upper() -- case insensitive
        if THUtils.debugFlags[realFlagId] == nil then
            THUtils.debugFlags[realFlagId] = false
        end
        return realFlagId
    end
end
function THUtils.setDebugFlag(flagId, isEnabled)
    if flagId == nil or flagId == "" then
        if isEnabled == true then
            THUtils.globalDebugFlag = true
        else
            THUtils.globalDebugFlag = false
        end
        for otherFlagId in pairs(THUtils.debugFlags) do
            THUtils.debugFlags[otherFlagId] = isEnabled
        end
        return nil, THUtils.globalDebugFlag
    end
    if type(flagId) == "table" and flagId.debugFlagId ~= nil then
        flagId = flagId.debugFlagId
    end
    local realFlagId = tostring(flagId):upper() -- case insensitive
    if THUtils.debugFlags[realFlagId] ~= nil then
        if isEnabled == true then
            THUtils.debugFlags[realFlagId] = true
        else
            THUtils.debugFlags[realFlagId] = false
        end
    else
        THUtils.warningMsg("Invalid debug flagId (%s)", flagId)
    end
    return realFlagId, THUtils.debugFlags[realFlagId]
end
function THUtils.getIsDebugEnabled(flagId, reqDebugLevelId)
    local reqDebugLevel = -math.huge
    if reqDebugLevelId ~= nil then
        reqDebugLevel = THUtils.getDebugLevelById(reqDebugLevelId) or math.huge
    end
    if THUtils.debugLevel >= reqDebugLevel then
        if flagId == nil or flagId == "" then
            return THUtils.globalDebugFlag == true
        end
        if type(flagId) == "table" and flagId.debugFlagId ~= nil then
            flagId = flagId.debugFlagId
        end
        local realFlagId = tostring(flagId):upper()
        if THUtils.debugFlags[realFlagId] == nil then
            THUtils.errorMsg(nil, THUtils.MSG.ARGUMENT_INVALID, "flagId", flagId)
        end
        return THUtils.debugFlags[realFlagId] == true
    end
    return false
end
function THUtils.setDebugLevel(debugLevelId)
    local debugLevel = THUtils.getDebugLevelById(debugLevelId)
    if debugLevel ~= nil then
        THUtils.debugLevel = debugLevel
    end
end
function THUtils.getDebugLevel()
    return THUtils.debugLevel or THUtils.DEBUG_LEVEL.NORMAL
end
function THUtils.getDebugLevelById(debugLevelId)
    local debugLevel = nil -- placeholder
    if type(debugLevelId) == "string" then
        debugLevel = THUtils.DEBUG_LEVEL[debugLevelId:upper()]
        if debugLevel == nil then
            THUtils.warningMsg("Invalid debug level (%s), valid values are:", debugLevelId)
            for allowedId in pairs(THUtils.DEBUG_LEVEL) do
                THUtils.displayMsg("- %s", allowedId)
            end
            THUtils.displayMsg("")
        end
    elseif debugLevelId ~= nil then
        debugLevel = tonumber(debugLevelId)
        if debugLevel == nil then
            THUtils.warningMsg(THUtils.MSG.ARGUMENT_INVALID, "debugLevelId", debugLevelId)
        end
    end
    return debugLevel
end
function THUtils.debugReset()
    THUtils.setDebugFlag(nil, false)
    THUtils.setDebugLevel(THUtils.DEBUG_LEVEL.NORMAL)
end
function THUtils.debugMsg(flagId, text, ...)
    if THUtils.getIsDebugEnabled(flagId) then
        if type(flagId) == "table" and flagId.debugFlagId ~= nil then
            flagId = flagId.debugFlagId
        end
        if flagId ~= nil and flagId ~= "" then
            if text ~= "" then
                text = string.format("[%s] %s", flagId, text)
            end
        end
        if text == "" then
            printf("")
        else
            THUtils.displayMsg("DEBUG: " .. text, ...)
        end
    end
end
function THUtils.debugWarningMsg(flagId, text, ...)
    if THUtils.getIsDebugEnabled(flagId) then
        if type(text) == "string" then
            text = "[WARNING] " .. text
        end
        THUtils.debugMsg(flagId, text, ...)
    end
end
function THUtils.debugErrorMsg(flagId, text, ...)
    if THUtils.getIsDebugEnabled(flagId) then
        if type(text) == "string" then
            text = "[ERROR] " .. text
        end
        THUtils.debugMsg(flagId, text, ...)
        printCallstack()
    end
end
function THUtils.getGlobalEnvironment()
    local globalEnv = _G
    local mt = getmetatable(_G)
    if mt ~= nil and mt.__index ~= nil then
        globalEnv = mt.__index
    end
    return globalEnv
end
function THUtils.pack(...)
    return {n = select("#", ...), ...}
end
function THUtils.unpack(target, index, endIndex)
    if endIndex == nil and target.n ~= nil then
        index = THUtils.getNoNil(index, 1)
        endIndex = target.n
    end
    return unpack(target, index, endIndex)
end
function THUtils.pcall(func, pa,pb,pc,pd,pe,pf,pg,ph,pi,pj,pk,pl,pm,pn,po,pp,pq,pr,ps,pt)
    local function errHandler(errMsg)
        errMsg = THUtils.getNoNil(errMsg, THUtils.MSG.INTERNAL_ERROR)
        if type(errMsg) == "string" then
            THUtils.errorMsg(nil, errMsg)
        else
            printCallstack()
        end
    end
    local function protectedFunc()
        return func(pa,pb,pc,pd,pe,pf,pg,ph,pi,pj,pk,pl,pm,pn,po,pp,pq,pr,ps,pt)
    end
    return xpcall(protectedFunc, errHandler)
end
function THUtils.call(func, ...)
    local function appendFunc(rSuccess, ret1, ...)
        if rSuccess == true then
            return ret1, ...
        end
    end
    return appendFunc(THUtils.pcall(func, ...))
end
function THUtils.toBoolean(value)
    if value == true or value == false then
        return value
    elseif type(value) == "string" then
        local upperVal = value:upper()
        if upperVal == "TRUE" then
            return true
        elseif upperVal == "FALSE" then
            return false
        end
    end
end
function THUtils.getNoNil(target, replacement, includeEmptyStrings)
    if target == nil or (target == "" and includeEmptyStrings == true) then
        return replacement
    end
    return target
end
function THUtils.getIsType(value, typeEntry)
    local valueType = type(value)
    if valueType == typeEntry then -- exact match (fastest)
        return true
    end
    local compareType = type(typeEntry)
    if compareType == "table" then
        if valueType == "table" and value.isa ~= nil and value:isa(typeEntry) then
            return true
        end
    elseif compareType == "string" then
        if string.find(typeEntry, "^" .. valueType .. " ")
            or string.find(typeEntry, " " .. valueType .. "$")
            or string.find(typeEntry, " " .. valueType .. " ")
        then
            return true
        end
    else
        THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "typeEntry", typeEntry) -- ! stop on error
    end
    return false
end
function THUtils.floor(value, precision)
    precision = precision or 0
    local factor = 10 ^ precision
    return math.floor(value * factor) / factor
end
function THUtils.ceil(value, precision)
    precision = precision or 0
    local factor = 10 ^ precision
    return math.ceil(value * factor) / factor
end
function THUtils.round(value, precision)
    precision = precision or 0
    local factor = 10 ^ precision
    return math.floor((value * factor) + 0.5) / factor
end
function THUtils.formatPercent(value, precision)
    precision = tonumber(precision) or 0
    precision = math.floor(precision)
    local dispValue = THUtils.round(value * 100, precision)
    local textValue = string.format("%0." .. tostring(precision) .. "f%%", dispValue)
    return textValue
end
function THUtils.getVector2Distance(sx,sz, ex,ez)
    return math.sqrt(((ex - sx) ^ 2) + ((ez - sz) ^ 2))
end
function THUtils.getVector3Distance(sx,sy,sz, ex,ey,ez)
    return math.sqrt(((ex - sx) ^ 2) + ((ey - sy) ^ 2) + ((ez - sz) ^ 2))
end
function THUtils.getFactorInRange(val1, val2, testVal, noClamp)
    local factor = (testVal - val1) / (val2 - val1)
    if noClamp ~= true then
        factor = MathUtil.clamp(factor, 0, 1)
    end
    return factor
end
function THUtils.getValueInRange(val1, val2, factor, noClamp)
    local newVal = val1 + (val2 - val1) * factor
    if noClamp ~= true then
        local minVal = math.min(val1, val2)
        local maxVal = math.max(val1, val2)
        newVal = MathUtil.clamp(newVal, minVal, maxVal)
    end
    return newVal
end
function THUtils.splitString(text, separator)
    local splitTable = string.split(text, separator)
    if splitTable == nil then
        splitTable = {}
    end
    return splitTable, #splitTable
end
function THUtils.properCase(text, removeSpaces)
    if text == nil then -- skip (pass through)
        return
    end
    local retText = text
    retText = string.gsub(retText, "^%l", string.upper)
    retText = string.gsub(retText, " %l", string.upper)
    if removeSpaces == true then
        retText = string.gsub(retText, " ", "")
    end
    return retText
end
function THUtils.snakeCase(text, toUpper)
    if text == nil then -- skip (pass through)
        return
    end
    local newText = string.gsub(text, " ", "_")
    newText = string.gsub(newText, "(%l)(%u)", "%1_%2")
    if toUpper == true then
        newText = newText:upper()
    end
    return newText
end
function THUtils.createClass(mainClass, baseClass, isProtected, ...)
    local mt = Class(mainClass, baseClass, ...)
    if mt ~= nil and isProtected == false then
        mt.__metatable = nil
    end
    return mt
end
function THUtils.clearTable(target)
    for key in pairs(target) do
        target[key] = nil
    end
    return target
end
function THUtils.copyTable(srcTable, depth)
    depth = THUtils.getNoNil(depth, 0)
    local newTable = {}
    if THUtils.argIsValid(type(srcTable) == "table", "srcTable", srcTable, true) then -- ! stop on error (prevent bad copy)
        for key, val in pairs(srcTable) do
            if type(val) == "table" and depth > 0 then
                newTable[key] = THUtils.copyTable(val, depth - 1)
            else
                newTable[key] = val
            end
        end
        local oldMt = getmetatable(srcTable)
        if oldMt ~= nil then
            setmetatable(newTable, oldMt)
        end
    end
    return newTable
end
function THUtils.createSubTable(parent, tableKey, clear, useRaw)
    local newTable = nil -- placeholder
    if useRaw == true then
        newTable = rawget(parent, tableKey)
    else
        newTable = parent[tableKey]
    end
    if newTable ~= nil then
        if clear == true then
            THUtils.clearTable(newTable)
        end
    else
        newTable = {}
        if useRaw == true then
            rawset(parent, tableKey, newTable)
        else
            parent[tableKey] = newTable
        end
    end
    return newTable
end
function THUtils.getTableLength(target)
    local numEntries = 0
    if target ~= nil then
        for _ in pairs(target) do
            numEntries = numEntries + 1
        end
    end
    return numEntries
end
function THUtils.getTableIsArray(target)
    local isArray = target[1] ~= nil
    if isArray then
        local numEntries = THUtils.getTableLength(target)
        if numEntries ~= #target then
            isArray = false
        end
    end
    return isArray
end
function THUtils.createDataTable(target, useRaw, customMt)
    if THUtils.argIsValid(type(target) == "table", "target", target, true) then -- ! stop on error (internal)
        local dataKey = THUtils.DATA_KEY
        local dataTable = nil
        if useRaw == true then
            dataTable = rawget(target, dataKey)
        else
            dataTable = target[dataKey]
        end
        if dataTable ~= nil and dataTable.parent ~= target then
            THUtils.errorMsg(true, "Data table target mismatch") -- ! stop on error (should never happen)
            return
        end
        if dataTable == nil then
            dataTable = {
                parent = target,
                isDataInitialized = {}
            }
            if useRaw == true then
                rawset(target, dataKey, dataTable)
            else
                target[dataKey] = dataTable
            end
        end
        if customMt ~= nil then
            local oldMt = getmetatable(dataTable)
            if oldMt == nil or oldMt ~= customMt then
                setmetatable(dataTable, customMt)
            end
        end
        return dataTable, target
    end
end
function THUtils.getDataTable(target, useRaw)
    if target ~= nil then -- nil values pass through
        local dataTable = nil -- placeholder
        local dataKey = THUtils.DATA_KEY
        if useRaw == true then
            dataTable = rawget(target, dataKey)
        else
            dataTable = target[dataKey]
        end
        return dataTable, target
    end
end
function THUtils.getSpecTable(object, specName)
    local specKey = string.format("spec_%s.%s", THUtils.MOD_NAME, specName)
    if object ~= nil then -- nil values pass through
        local specTable = object[specKey]
        local objectData = THUtils.getDataTable(object)
        return specTable, objectData, object
    end
    return nil, nil, object
end
function THUtils.makeSelfCallback(targetFunc, newSelf)
    if newSelf == nil then
        THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "newSelf", newSelf, true) -- ! stop on error
    else
        local callbackFunc = function(...)
            return targetFunc(newSelf, ...)
        end
        return callbackFunc
    end
end
function THUtils.registerFunction(target, funcName, newFunc)
    if THUtils.argIsValid(type(target) == "table", "target", target, true) -- ! stop on error
        and THUtils.argIsValid(type(funcName) == "string", "funcName", funcName, true) -- ! stop on error
        and THUtils.argIsValid(type(newFunc) == "function", "newFunc", newFunc, true) -- ! stop on error
    then
        local oldFunc = rawget(target, funcName)
        if oldFunc ~= nil then
            THUtils.errorMsg(true, "Function %q already registered", funcName) -- ! stop on error
            return oldFunc
        end
        local callbackFunc = function(...)
            return newFunc(...)
        end
        rawset(target, funcName, callbackFunc)
        return callbackFunc
    end
end
function THUtils.setFunctionHook(srcClass, srcFuncKey, tgtClass, tgtFuncKey, useSelf, extraArg)
    useSelf = THUtils.getNoNil(useSelf, true)
    if type(srcClass) == "string" then
        local globalEnv = THUtils.getGlobalEnvironment()
        srcClass = globalEnv[srcClass]
    end
    if type(tgtClass) == "string" then
        tgtClass = _G[tgtClass]
    end
    if tgtFuncKey == nil then
        if type(srcFuncKey) == "string" then
            tgtFuncKey = string.format("hook_%s", srcFuncKey)
        end
    end
    local selfType = type(useSelf)
    local newSelf = nil -- placeholder
    if selfType == "table" then
        newSelf = useSelf
        useSelf = true
    elseif selfType ~= "boolean" then
        THUtils.errorMsg(true, THUtils.MSG.INVALID_ARGUMENT, "useSelf", useSelf) -- ! stop on error
        return
    end
    if THUtils.argIsValid(type(srcClass) == "table", "srcClass", srcClass, true) -- ! stop on error
        and THUtils.argIsValid(srcFuncKey ~= nil, "srcFuncKey", srcFuncKey, true) -- ! stop on error
        and THUtils.argIsValid(type(tgtClass) == "table", "tgtClass", tgtClass, true) -- ! stop on error
        and THUtils.argIsValid(tgtFuncKey ~= nil, "tgtFuncKey", tgtFuncKey, true) -- ! stop on error
    then
        local srcFunc = srcClass[srcFuncKey]
        local tgtFunc = tgtClass[tgtFuncKey]
        if THUtils.assert(type(srcFunc) == "function", true, "Invalid source function (%s)", srcFuncKey) -- ! stop on error
            and THUtils.assert(type(tgtFunc) == "function", true, "Invalid target function (%s)", tgtFuncKey) -- ! stop on error
        then
            local function retFunc(p1, ...)
                if useSelf == true then
                    if newSelf ~= nil then
                        if extraArg ~= nil then
                            return tgtFunc(newSelf, srcFunc, extraArg, p1, ...)
                        else
                            return tgtFunc(newSelf, srcFunc, p1, ...)
                        end
                    elseif extraArg ~= nil then
                        return tgtFunc(p1, srcFunc, extraArg, ...)
                    else
                        return tgtFunc(p1, srcFunc, ...)
                    end
                elseif extraArg ~= nil then
                    return tgtFunc(srcFunc, extraArg, p1, ...)
                else
                    return tgtFunc(srcFunc, p1, ...)
                end
            end
            rawset(srcClass, srcFuncKey, retFunc) -- bypass metamethods
            return retFunc
        end
    end
end
function THUtils.createFunctionCacheId(arg1, arg2)
    local prefix = tostring(arg1)
    local suffix = tostring(arg2)
    THUtils.msgOnTrue(arg1 == nil, nil, THUtils.MSG.ARGUMENT_INVALID, "arg1", arg1)
    THUtils.msgOnTrue(arg2 == nil, nil, THUtils.MSG.ARGUMENT_INVALID, "arg2", arg2)
    return string.format("%s|%s", prefix, suffix)
end
function THUtils.setCachedHook(cacheId, target, funcKey, newFunc, extraArg, verbose)
    if cacheId ~= nil then
        cacheId = tostring(cacheId)
    end
    if type(target) == "string" then
        local globalEnv = THUtils.getGlobalEnvironment()
        target = globalEnv[target]
    end
    if THUtils.argIsValid(cacheId ~= nil, "cacheId", cacheId, true) -- ! stop on error
        and THUtils.argIsValid(type(target) == "table", "target", target, true) -- ! stop on error
        and THUtils.argIsValid(funcKey ~= nil, "funcKey", funcKey, true) -- ! stop on error
        and THUtils.argIsValid(type(newFunc) == "function", "newFunc", newFunc, true) -- ! stop on error
    then
        local oldFunc = target[funcKey]
        local rawFunc = rawget(target, funcKey)
        if THUtils.assert(type(oldFunc) == "function", true, "Invalid source function (%s)", funcKey) then -- ! stop on error
            local cacheData = THUtils.functionCache[cacheId]
            local entryId = tostring(oldFunc)
            if cacheData == nil then
                cacheData = {}
                THUtils.functionCache[cacheId] = cacheData
            end
            local cacheEntry = cacheData[entryId]
            if cacheEntry == nil then
                cacheEntry = {
                    cacheId = cacheId,
                    entryId = entryId,
                    target = target,
                    funcKey = funcKey,
                    oldFunc = oldFunc,
                    newFunc = newFunc,
                    rawFunc = rawFunc
                }
                rawset(target, funcKey, function(...)
                    if extraArg ~= nil then
                        return newFunc(oldFunc, extraArg, ...)
                    else
                        return newFunc(oldFunc, ...)
                    end
                end)
                cacheData[entryId] = cacheEntry
            elseif verbose == true then
                THUtils.errorMsg(true, "Temporary hook (%s:%s) already exists", cacheId, funcKey) -- ! stop on error
            end
            return cacheEntry
        end
    end
end
function THUtils.restoreFunction(cacheId, target, funcKey)
    if cacheId ~= nil then
        cacheId = tostring(cacheId)
    end
    if type(target) == "string" then
        local globalEnv = THUtils.getGlobalEnvironment()
        target = globalEnv[target]
    end
    local cacheData = THUtils.functionCache[cacheId]
    if THUtils.argIsValid(cacheData ~= nil, "cacheId", cacheId)
        and THUtils.argIsValid(type(target) == "table", "target", target)
        and THUtils.argIsValid(funcKey ~= nil, "funcKey", funcKey)
    then
        local oldFunc = target[funcKey]
        if THUtils.argIsValid(oldFunc ~= nil, "funcKey", funcKey) then
            local entryId = tostring(oldFunc)
            local cacheEntry = cacheData[entryId]
            if cacheEntry ~= nil then
                rawset(cacheEntry.target, cacheEntry.funcKey, cacheEntry.rawFunc)
                cacheData[entryId] = nil
                return true
            end
        end
    end
    return false
end
function THUtils.restoreAllFunctions(cacheId)
    if THUtils.argIsValid(cacheId ~= nil, "cacheId", cacheId) then
        local cacheData = THUtils.functionCache[cacheId]
        if cacheData ~= nil then
            for _, cacheEntry in pairs(cacheData) do
                rawset(cacheEntry.target, cacheEntry.funcKey, cacheEntry.rawFunc)
            end
            THUtils.clearTable(cacheData)
            return true
        end
    end
    return false
end
function THUtils.registerXMLPath(schema, xmlValueType, xmlPath, xmlKeys, ...)
    local xmlKeyType = type(xmlKeys)
    local xmlKeyList = nil -- placeholder
    if xmlKeyType == "string" and xmlKeys ~= "" then
        if string.find(xmlKeys, " ") then
            xmlKeyList = THUtils.splitString(xmlKeys, " ")
        else
            schema:register(xmlValueType, xmlPath .. xmlKeys, ...)
            return
        end
    elseif xmlKeyType == "table" then
        xmlKeyList = xmlKeys
    else
        THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "xmlKeys", xmlKeys) -- ! stop on error
        return
    end
    if xmlKeyList ~= nil and #xmlKeyList > 0 then
        for _, xmlKey in ipairs(xmlKeyList) do
            if xmlKey ~= "" then
                schema:register(xmlValueType, xmlPath .. xmlKey, ...)
            end
        end
    end
end
function THUtils.getXMLValue(xmlFile, basePath, xmlKeys, defaultValue, ...)
    local function protectedFunc(...)
        local xmlKeyType = type(xmlKeys)
        local xmlKeyList = nil -- placeholder
        if xmlKeyType == "string" then
            if string.find(xmlKeys, " ") then
                xmlKeyList = THUtils.splitString(xmlKeys, " ")
            else
                return xmlFile:getValue(basePath .. xmlKeys, defaultValue, ...)
            end
        elseif xmlKeyType == "table" then
            xmlKeyList = xmlKeys
        else
            THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "xmlKeys", xmlKeys) -- ! stop on error
            return
        end
        if xmlKeyList ~= nil and #xmlKeyList > 0 then
            for i, xmlKey in ipairs(xmlKeyList) do
                local xmlKeyFull = string.format("%s%s", basePath, xmlKey)
                local retValue, ra,rb,rc,rd,re,rf,rg,rh,ri,rj = xmlFile:getValue(xmlKeyFull, nil, ...)
                if retValue ~= nil then
                    if i > 1 then
                        local msgKey1 = xmlKeyFull
                        local msgKey2 = string.gsub(xmlKeyList[1], "^#", "")
                        THUtils.warningMsg("Outdated xml key: [%s] Please use %q instead", msgKey1, msgKey2)
                    end
                    return retValue, ra,rb,rc,rd,re,rf,rg,rh,ri,rj
                end
            end
        end
    end
    local function appendFunc(rSuccess, rValue, ...) -- ensures default value is always returned
        if rSuccess and rValue ~= nil then
            return rValue, ...
        else
            return defaultValue
        end
    end
    return appendFunc(THUtils.pcall(protectedFunc, ...))
end
function THUtils.getDataXMLFile(targetXMLFilename, dataKey, baseKey, xmlSchema)
    local targetXMLFile = XMLFile.loadIfExists(dataKey .. "FilenameXML", targetXMLFilename)
    baseKey = THUtils.getNoNil(baseKey, "map")
    if targetXMLFile ~= nil then
        local filenameKey = string.format("%s.%s#filename", baseKey, dataKey)
        local dataXMLFilename = targetXMLFile:getString(filenameKey)
        if dataXMLFilename ~= nil then
            local dataXMLFile = XMLFile.loadIfExists(dataKey, dataXMLFilename, xmlSchema)
            return dataXMLFile
        end
    end
end
function THUtils.getNormalizedValues(uiScale, x,y)
    local normX, normY = 0, 0
    if x ~= nil then
        normX = x * uiScale * g_aspectScaleX / g_referenceScreenWidth
    end
    if y ~= nil then
        normY = y * uiScale * g_aspectScaleY / g_referenceScreenHeight
    end
    return normX, normY
end
function THUtils.getElementByProfileName(baseElement, profileName, recursive)
    if profileName == nil then
        return
    end
    if profileName == baseElement.profile then
        return baseElement
    end
    local elementList = baseElement.elements
    if elementList and #elementList > 0 then
        for idx = 1, #elementList do
            local element = elementList[idx]
            if recursive then
                local foundElement = THUtils.getElementByProfileName(element, profileName, recursive)
                if foundElement ~= nil then
                    return foundElement
                end
            elseif profileName == element.profile then
                return element
            end
        end
    end
end