-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved


THSettingsManager = {}
local THSettingsManager_mt = THUtils.createClass(THSettingsManager, THAbstractSystem)
local THSettingsManager_dataKey = g_thMain.xmlDataKey
THSettingsManager.SETTING_NUM_BITS = 16
THSettingsManager.SETTING_VALUE_NUM_BITS = 32
THSettingsManager.ELEMENT_TYPE = {
    CHECKED = 1,
    MULTI_OPTION = 2
}
THSetting = {}
local THSetting_mt = THUtils.createClass(THSetting)
function THSettingsManager.initXMLSchemaSettings(xmlSchema, xmlPath)
    THSettingsManager:superClass().initXMLSchemaSettings(xmlSchema, xmlPath)
    local settingPath = xmlPath .. ".settings.setting(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, settingPath, "#name", "Setting name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, settingPath, "#valueIndex", "Currently selected value")
end
function THSettingsManager.new(xmlDataKey, customMt)
    xmlDataKey = xmlDataKey or THSettingsManager_dataKey
    customMt = customMt or THSettingsManager_mt
    local self = THAbstractSystem.new(xmlDataKey, customMt)
    self.settingCategories = {
        byId = {},
        byIndex = {},
        idToIndex = {}
    }
    self.SETTING_CATEGORY = self.settingCategories.idToIndex
    self.settings = {
        byId = {},
        byIndex = {},
        idToIndex = {}
    }
    self.SETTING = self.settings.idToIndex
    self.areSettingsDirty = false
    self.settingCategoryElements = {}
    self.settingElements = {
        byIndex = {},
        byType = {}
    }
    self.elementTypeIndexToId = {}
    for elementTypeId, elementTypeIndex in pairs(THSettingsManager.ELEMENT_TYPE) do
        self.elementTypeIndexToId[elementTypeIndex] = elementTypeId
        self.settingElements.byType[elementTypeIndex] = {}
    end
    self.isMenuInitialized = {}
    g_thMain:addEventListener(self)
    return self
end
function THSettingsManager:loadModSettingsFromXMLFile(xmlFile, xmlKey)
    local loadSuccess = THSettingsManager:superClass().loadModSettingsFromXMLFile(self, xmlFile, xmlKey)
    if loadSuccess then
        xmlFile:iterate(xmlKey .. ".settings.setting", function(_, pSettingKey)
            local settingName = THUtils.getXMLValue(xmlFile, pSettingKey, "#name")
            local valueIndex = THUtils.getXMLValue(xmlFile, pSettingKey, "#valueIndex", -1)
            local settingInfo = self:getSetting(settingName)
            if settingInfo == nil then
                THUtils.xmlLoadMsg(pSettingKey, "name", settingName, false, THUtils.XMLMSG.INVALID_VALUE)
            elseif valueIndex <= 0 then
                THUtils.xmlLoadMsg(pSettingKey, "valueIndex", valueIndex, true, THUtils.XMLMSG.INVALID_VALUE)
            else
                if self.isServer or not settingInfo.isMultiplayer then
                    local newValue = settingInfo:setValueByIndex(valueIndex, true)
                    if newValue ~= nil then
                        if settingInfo.defaultValue <= 0 then
                            settingInfo.defaultValue = valueIndex
                        end
                    end
                end
            end
        end)
    end
    return loadSuccess
end
function THSettingsManager:saveModSettingsToXMLFile(xmlFile, xmlKey)
    local saveSuccess = THSettingsManager:superClass().saveModSettingsToXMLFile(self, xmlFile, xmlKey)
    if saveSuccess then
        local settingsArray, numSettings = self:getSettings(true)
        if numSettings > 0 then
            THUtils.call(function()
                local settingsXMLKey = string.format("%s.settings", xmlKey)
                local settingXMLIndex = 0
                xmlFile:removeProperty(settingsXMLKey)
                for settingIndex = 1, numSettings do
                    local settingInfo = settingsArray[settingIndex]
                    local settingXMLKey = string.format("%s.setting(%d)", settingsXMLKey, settingXMLIndex)
                    local valueIndex = settingInfo.changedValue
                    if self.isServer or not settingInfo.isMultiplayer
                        and valueIndex > 0 and settingInfo.values[valueIndex] ~= nil
                    then
                        xmlFile:setString(settingXMLKey .. "#name", settingInfo.name)
                        xmlFile:setInt(settingXMLKey .. "#valueIndex", valueIndex)
                        settingInfo.savedValue = valueIndex
                        settingXMLIndex = settingXMLIndex + 1
                    end
                end
            end)
        end
    end
    return saveSuccess
end
function THSettingsManager:onInitialize()
    if not self.isInitFinished then
        THSettingsManager:superClass().onInitialize(self)
        self.xmlSchemaSettings = g_thMain.xmlSchemaSettings
        self.dirtyFlag = self:getNextDirtyFlag()
        THUtils.setFunctionHook("InGameMenuGeneralSettingsFrame", "onFrameOpen", THSettingsManager, nil, self)
        self:loadModSettings()
        self.isInitFinished = true
    end
end
function THSettingsManager:onUpdateTick(dt)
    THSettingsManager:superClass().onUpdateTick(self, dt)
    if self.areSettingsDirty then
        self:saveModSettings()
        self.areSettingsDirty = false
    end
end
function THSettingsManager:onClickSettingElement(state, element, ...)
    local elementData = THUtils.getDataTable(element)
    if elementData ~= nil and elementData.setting ~= nil then
        local settingInfo = elementData.setting
        local isMasterUser = self:getIsMasterUser()
        if (self.isServer or not settingInfo.isMultiplayer)
            and (isMasterUser or not settingInfo.isAdminOnly)
        then
            settingInfo:setValueByIndex(state)
        end
    end
end
function THSettingsManager:addSettingCategory(name, title, toolTip, customEnv)
    toolTip = THUtils.getNoNil(toolTip, "")
    customEnv = THUtils.getNoNil(customEnv, self.modName)
    if THUtils.argIsValid(type(name) == "string" and name ~= "", "name", name, true) -- ! stop on error
        and THUtils.argIsValid(type(title) == "string", "title", title, true) -- ! stop on error
        and THUtils.argIsValid(type(toolTip) == "string", "toolTip", toolTip, true) -- ! stop on error
        and THUtils.argIsValid(type(customEnv) == "string", "customEnv", customEnv, true) -- ! stop on error
    then
        local i18n = self.i18n
        local categoryId = name:upper()
        local categoryInfo = self.settingCategories.byId[categoryId]
        if categoryInfo ~= nil then
            THUtils.warningMsg("Setting category (%s) already exists", categoryInfo.name)
        else
            categoryInfo = {
                id = categoryId,
                name = name,
                index = #self.settingCategories.byIndex + 1,
                title = i18n:convertText(title, customEnv),
                toolTip = i18n:convertText(toolTip, customEnv),
                settings = {}
            }
            local idToIndexKey = THUtils.snakeCase(categoryInfo.name, true)
            self.settingCategories.byIndex[categoryInfo.index] = categoryInfo
            self.settingCategories.byId[categoryInfo.id] = categoryInfo
            self.settingCategories.idToIndex[idToIndexKey] = categoryInfo.index
        end
        return categoryInfo
    end
end
function THSettingsManager:addSetting(name, category, title, toolTip, elementType, customEnv, restartRequired, isMultiplayer, isAdminOnly)
    toolTip = THUtils.getNoNil(toolTip, "")
    customEnv = THUtils.getNoNil(customEnv, self.modName)
    if THUtils.argIsValid(type(name) == "string" and name ~= "", "name", name, true) -- ! stop on error
        and THUtils.argIsValid(type(title) == "string", "title", title, true) -- ! stop on error
        and THUtils.argIsValid(type(toolTip) == "string", "toolTip", toolTip, true) -- ! stop on error
        and THUtils.argIsValid(type(customEnv) == "string", "customEnv", customEnv, true) -- ! stop on error
    then
        local i18n = self.i18n
        local settingId = name:upper()
        local categoryInfo = self:getSettingCategory(category)
        if categoryInfo == nil then
            THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "category", category) -- ! stop on error
        elseif elementType == nil or self.elementTypeIndexToId[elementType] == nil then
            THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "elementType", elementType) -- ! stop on error
        else
            local settingInfo = self.settings.byId[settingId]
            if settingInfo ~= nil then
                THUtils.warningMsg("Setting (%s) already exists", settingInfo.name)
            else
                settingInfo = {
                    id = settingId,
                    name = name,
                    index = #self.settings.byIndex + 1,
                    parent = self,
                    customEnv = customEnv,
                    title = i18n:convertText(title, customEnv),
                    toolTip = i18n:convertText(toolTip, customEnv),
                    elementType = elementType,
                    categoryIndex = categoryInfo.index,
                    isClient = self.isClient,
                    isServer = self.isServer,
                    isMultiplayer = isMultiplayer == true,
                    isAdminOnly = isAdminOnly == true,
                    restartRequired = restartRequired == true,
                    isLocked = false,
                    values = {},
                    valueMapping = {},
                    valueTexts = {},
                    currentValue = 0,
                    changedValue = 0,
                    savedValue = 0,
                    defaultValue = 0
                }
                setmetatable(settingInfo, THSetting_mt)
                if settingInfo.elementType == THSettingsManager.ELEMENT_TYPE.CHECKED then
                    settingInfo:addValue(false, "$l10n_ui_off")
                    settingInfo:addValue(true, "$l10n_ui_on")
                    settingInfo.isLocked = true
                end
                settingInfo.messageType = nextMessageTypeId()
                local idToIndexKey = THUtils.snakeCase(settingInfo.name, true)
                self.settings.byId[settingInfo.id] = settingInfo
                self.settings.byIndex[settingInfo.index] = settingInfo
                self.settings.idToIndex[idToIndexKey] = settingInfo.index
                table.insert(categoryInfo.settings, settingInfo.index)
            end
            return settingInfo
        end
    end
end
function THSettingsManager:addSettingChangedListener(setting, target, callbackFunc)
    local settingInfo = self:getSetting(setting)
    if settingInfo ~= nil
        and THUtils.argIsValid(type(target) == "table", "target", target, true) -- ! stop on error
        and THUtils.argIsValid(callbackFunc == nil or type(callbackFunc) == "function", "callbackFunc", callbackFunc, true) -- ! stop on error
    then
        if callbackFunc == nil then
            callbackFunc = target.onSettingChanged
            if callbackFunc == nil then
                THUtils.errorMsg(true, "default callback function onSettingChanged does not exist") -- ! stop on error
                return
            end
        end
        self.messageCenter:subscribe(settingInfo.messageType, callbackFunc, target)
    end
end
function THSettingsManager:removeSettingChangedListener(setting, target)
    local settingInfo = self:getSetting(setting)
    if settingInfo ~= nil then
        self.messageCenter:unsubscribe(settingInfo.messageType, target)
    end
end
function THSettingsManager:getSettingCategory(category)
    local categoryInfo = nil -- placeholder
    local valType = type(category)
    if valType == "string" then
        local categoryId = category:upper()
        categoryInfo = self.settingCategories.byId[categoryId]
        if categoryInfo == nil then
            local categoryIndex = self.SETTING_CATEGORY[categoryId]
            if categoryIndex ~= nil then
                categoryInfo = self.settingCategories.byIndex[categoryIndex]
            end
        end
    elseif valType == "number" then
        categoryInfo = self.settingCategories.byIndex[category]
    elseif valType == "table" and category.id ~= nil then
        categoryInfo = self.settingCategories.byId[category.id]
        if categoryInfo ~= category then
            categoryInfo = nil
        end
    elseif category ~= nil then -- nil values pass though
        THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "category", category) -- ! stop on error
    end
    return categoryInfo
end
function THSettingsManager:getSettingCategories(byIndex)
    local numCategories = #self.settingCategories.byIndex
    if byIndex == true then
        return self.settingCategories.byIndex, numCategories
    else
        return self.settingCategories.byId, numCategories
    end
end
function THSettingsManager:getSetting(setting)
    local settingInfo = nil -- placeholder
    local valType = type(setting)
    if valType == "string" then
        local settingId = setting:upper()
        settingInfo = self.settings.byId[settingId]
        if settingInfo == nil then
            local settingIndex = self.SETTING[settingId]
            if settingIndex ~= nil then
                settingInfo = self.settings.byIndex[settingIndex]
            end
        end
    elseif valType == "number" then
        settingInfo = self.settings.byIndex[setting]
    elseif valType == "table" and setting.id ~= nil then
        settingInfo = self.settings.byId[setting.id]
        if settingInfo ~= setting then
            settingInfo = nil
        end
    elseif setting ~= nil then -- nil values pass through
        THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "setting", setting) -- ! stop on error
    end
    return settingInfo
end
function THSettingsManager:getSettings(byIndex)
    local numSettings = #self.settings.byIndex
    if byIndex == true then
        return self.settings.byIndex, numSettings
    else
        return self.settings.byId, numSettings
    end
end
function THSettingsManager:getSettingValue(setting)
    local settingInfo = self:getSetting(setting)
    if settingInfo ~= nil then
        local settingValue, valueIndex = settingInfo:getCurrentValue()
        return settingValue, valueIndex, settingInfo
    end
end
function THSettingsManager:initializeSettingElements(menuFrame)
    if not self.isMenuInitialized[menuFrame] then
        local boxLayoutElement = menuFrame.boxLayout
        local baseHeaderElement = THUtils.getElementByProfileName(boxLayoutElement, "settingsMenuSubtitle")
        local baseCheckedOptionElement = menuFrame.checkUseMiles
        local baseMultiOptionElement = menuFrame.multiMoneyUnit
        if boxLayoutElement ~= nil and baseHeaderElement ~= nil and baseMultiOptionElement ~= nil then
            local categoriesArray, numCategories = self:getSettingCategories(true)
            if numCategories > 0 then
                for categoryIndex = 1, numCategories do
                    local categoryInfo = categoriesArray[categoryIndex]
                    local numSettings = #categoryInfo.settings
                    if numSettings > 0 then
                        local categoryElement = baseHeaderElement:clone(boxLayoutElement)
                        categoryElement:setText(categoryInfo.title)
                        categoryElement:reloadFocusHandling(true)
                        local categoryElementData = THUtils.createDataTable(categoryElement)
                        categoryElementData.settingCategory = categoryInfo
                        table.insert(self.settingCategoryElements, categoryElement)
                        for i = 1, numSettings do
                            local settingIndex = categoryInfo.settings[i]
                            local settingInfo = self:getSetting(settingIndex)
                            local settingElement = nil -- placeholder
                            if settingInfo.defaultValue <= 0 then
                                THUtils.warningMsg("Setting (%s): No default value specified, skipping...")
                            elseif settingInfo.elementType == THSettingsManager.ELEMENT_TYPE.CHECKED then
                                settingElement = baseCheckedOptionElement:clone(boxLayoutElement)
                            elseif settingInfo.elementType == THSettingsManager.ELEMENT_TYPE.MULTI_OPTION then
                                settingElement = baseMultiOptionElement:clone(boxLayoutElement)
                            end
                            if settingElement == nil then
                                THUtils.warningMsg("Could not create menu element for setting (%s)", settingInfo.name)
                            else
                                settingElement.onClickCallback = function(pSelf, ...)
                                    return self:onClickSettingElement(...)
                                end
                                local titleElement = settingElement.elements[4]
                                local toolTipElement = settingElement.elements[6]
                                local settingElementData = THUtils.createDataTable(settingElement)
                                settingElementData.setting = settingInfo
                                settingElementData.titleElement = titleElement
                                settingElementData.toolTipElement = toolTipElement
                                titleElement:setText(settingInfo.title)
                                toolTipElement:setText(settingInfo.toolTip)
                                settingElement:setTexts(settingInfo.valueTexts)
                                settingElement:reloadFocusHandling(true)
                                table.insert(self.settingElements.byIndex, settingElement)
                                table.insert(self.settingElements.byType[settingInfo.elementType], settingElement)
                            end
                        end
                    end
                end
            end
        end
        self.isMenuInitialized[menuFrame] = true
    end
    self:updateSettingElements()
end
function THSettingsManager:updateSettingElements()
    local elementsArray, numElements = self:getSettingElements()
    if numElements > 0 then
        local isMasterUser = self:getIsMasterUser()
        for elementIndex = 1, numElements do
            local elementData, element = THUtils.getDataTable(elementsArray[elementIndex])
            if elementData ~= nil then
                local settingInfo = elementData.setting
                local settingValue, valueIndex = settingInfo:getLastChangedValue()
                local isElementDisabled = false
                if settingValue == nil or valueIndex <= 0 then
                    settingValue, valueIndex = settingInfo:getDefaultValue()
                end
                if settingValue == nil or valueIndex <= 0 then
                    isElementDisabled = true
                else
                    element:setState(valueIndex)
                    if (settingInfo.isMultiplayer and not self.isServer)
                        or (settingInfo.isAdminOnly and not isMasterUser)
                    then
                        isElementDisabled = true
                    end
                end
                if element:getIsDisabled() ~= isElementDisabled then
                    element:setDisabled(isElementDisabled)
                end
            end
        end
    end
end
function THSettingsManager:getSettingElements(elementTypeId)
    local elementType = nil -- placeholder
    if elementTypeId ~= nil then
        local idType = type(elementTypeId)
        if idType == "number" and self.elementTypeIndexToId[elementTypeId] ~= nil then
            elementType = elementTypeId
        elseif idType == "string" then
            elementType = THSettingsManager.ELEMENT_TYPE[elementTypeId:upper()]
        end
        if elementType == nil then
            THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "elementTypeId", elementTypeId) -- ! stop on error
            return {}, 0
        end
    end
    if elementType ~= nil then
        return self.settingElements.byType[elementType], #self.settingElements.byType[elementType]
    else
        return self.settingElements.byIndex, #self.settingElements.byIndex
    end
end
function THSettingsManager:hook_onFrameOpen(superFunc, menuFrame, ...)
    THUtils.call(function()
        if menuFrame ~= nil and menuFrame.name == "ingameMenuGameSettingsGeneral" then
            self:initializeSettingElements(menuFrame)
        end
    end)
    return superFunc(menuFrame, ...)
end
function THSetting:getCurrentValue()
    return self.values[self.currentValue], self.currentValue
end
function THSetting:getLastChangedValue()
    return self.values[self.changedValue], self.changedValue
end
function THSetting:getDefaultValue()
    return self.values[self.defaultValue], self.defaultValue
end
function THSetting:addValue(value, title)
    if THUtils.argIsValid(value ~= nil, "value", value, true) -- ! stop on error
        and THUtils.argIsValid(type(title) == "string" and title ~= "", "title", title, true) -- ! stop on error
    then
        local i18n = self.parent.i18n
        local valType = type(value)
        local valueId = value
        if valType == "string" then
            valueId = value:upper()
        end
        if self.valueMapping[valueId] ~= nil then
            THUtils.warningMsg("Value (%s) has already been added to setting (%s)", value, self.name)
        else
            local valueIndex = #self.values + 1
            self.values[valueIndex] = value
            self.valueTexts[valueIndex] = i18n:convertText(title, self.customEnv)
            self.valueMapping[valueId] = valueIndex
            if self.changedValue <= 0 then
                self.changedValue = 1
            end
            if self.currentValue <= 0 then
                self.currentValue = 1
            end
            if self.defaultValue <= 0 then
                self.defaultValue = 1
            end
            if self.savedValue <= 0 then
                self.savedValue = 1
            end
        end
    end
end
function THSetting:setValue(value, noEventSend, updateDefault)
    local valType = type(value)
    local valueIndex = -1
    if valType == "string" then
        valueIndex = self.valueMapping[value:upper()]
    elseif value ~= nil then
        valueIndex = self.valueMapping[value]
    else
        THUtils.errorMsg(true, THUtils.MSG.ARGUMENT_INVALID, "value", value) -- ! stop on error
    end
    if valueIndex ~= nil and valueIndex > 0 then
        local newValue = self:setValueByIndex(valueIndex, noEventSend, updateDefault)
        if newValue ~= nil then
            return valueIndex, newValue
        end
    end
    return valueIndex
end
function THSetting:setValueByIndex(valueIndex, noEventSend, updateDefault)
    local messageCenter = self.parent.messageCenter
    updateDefault = THUtils.getNoNil(updateDefault, false)
    if THUtils.argIsValid(type(valueIndex) == "number", "valueIndex", valueIndex, true) -- ! stop on error
        and THUtils.argIsValid(type(updateDefault) == "boolean", "updateDefault", updateDefault, true) -- ! stop on error
    then
        local isMasterUser = self.parent:getIsMasterUser()
        local newValue = self.values[valueIndex]
        if newValue ~= nil then
            if self.isServer or not self.isMultiplayer or noEventSend == true then
                self.changedValue = valueIndex
                if not self.restartRequired then
                    self.currentValue = valueIndex
                end
                if updateDefault == true then
                    self.defaultValue = valueIndex
                end
                messageCenter:publish(self.messageType, self, newValue, valueIndex)
                if noEventSend ~= true then
                    if self.isMultiplayer and self.isServer then
                        self.parent:raiseDirtyFlags(self.parent.dirtyFlag)
                    end
                    self.parent.areSettingsDirty = true
                    self.parent:raiseActive()
                end
                return newValue, valueIndex
            elseif not self.isAdminOnly or isMasterUser then
                if THSetSettingValueEvent.sendEvent(self.index, valueIndex, updateDefault) then
                    return newValue, valueIndex
                end
            end
        end
    end
    return nil, -1
end
THSetSettingValueEvent = {}
local THSetSettingValueEvent_mt = THUtils.createClass(THSetSettingValueEvent, Event)
InitEventClass(THSetSettingValueEvent, "THSetSettingValueEvent")
function THSetSettingValueEvent.emptyNew()
    return Event.new(THSetSettingValueEvent_mt)
end
function THSetSettingValueEvent.new(settingIndex, valueIndex, updateDefault)
    local self = THSetSettingValueEvent.emptyNew()
    self.settingIndex = settingIndex
    self.valueIndex = valueIndex
    self.updateDefault = updateDefault
    return self
end
function THSetSettingValueEvent:readStream(streamId, connection)
    if not connection:getIsServer() then -- server reading from client
        self.settingIndex = streamReadUIntN(streamId, THSettingsManager.SETTING_NUM_BITS)
        self.valueIndex = streamReadUIntN(streamId, THSettingsManager.SETTING_VALUE_NUM_BITS)
        self.updateDefault = streamReadBool(streamId)
        self:run(connection)
    else
        THUtils.errorMsg(nil, "THSetSettingValueEvent should not be sent from the server!")
    end
end
function THSetSettingValueEvent:writeStream(streamId, connection)
    if connection:getIsServer() then -- client sending to server
        streamWriteUIntN(streamId, self.settingIndex, THSettingsManager.SETTING_NUM_BITS)
        streamWriteUIntN(streamId, self.valueIndex, THSettingsManager.SETTING_VALUE_NUM_BITS)
        streamWriteBool(streamId, self.updateDefault)
    else
        THUtils.errorMsg(nil, "THSetSettingValueEvent should not be sent from the server!")
    end
end
function THSetSettingValueEvent:run(connection)
    if not connection:getIsServer() then -- server reading from client
        local settingsManager = g_thMain.settingsManager
        local settingInfo = settingsManager:getSetting(self.settingIndex)
        if settingInfo == nil then
            THUtils.errorMsg(true, "Invalid setting index (%s) sent to server!", self.settingIndex) -- ! stop on error
        elseif not settingInfo.isMultiplayer then
            THUtils.errorMsg(true, "Non multiplayer setting (%s) should not be synchronized!", self.settingIndex) -- ! stop on error
        else
            settingInfo:setValueByIndex(self.valueIndex, nil, self.updateDefault)
        end
    else
        THUtils.errorMsg(nil, "THSetSettingEvent should never run on clients!")
    end
end
function THSetSettingValueEvent.sendEvent(settingIndex, valueIndex, updateDefault)
    if THUtils.argIsValid(type(settingIndex) == "number" and settingIndex > 0, "settingIndex", settingIndex, true) -- ! stop on error
        and THUtils.argIsValid(type(valueIndex) == "number" and valueIndex > 0, "valueIndex", valueIndex, true) -- ! stop on error
        and THUtils.argIsValid(type(updateDefault) == "boolean", "updateDefault", updateDefault, true) -- ! stop on error
    then
        if not g_thMain.isServer and g_client ~= nil then
            local newEvent = THSetSettingValueEvent.new(settingIndex, valueIndex, updateDefault)
            if newEvent ~= nil then
                g_client:getServerConnection():sendEvent(newEvent)
                return true
            end
        end
    end
    return false
end